from logging import getLogger

import numpy as np

from . import parser

logger = getLogger(__name__)

def wer_tabular(hyp: list[str], ref: list[str]):
    table = np.zeros((len(hyp) + 1, len(ref) + 1), dtype=np.uint8)
    table[:, 0] = np.arange(len(hyp) + 1)
    table[0, :] = np.arange(len(ref) + 1)

    for i in range(1, len(hyp) + 1):
        for j in range(1, len(ref) + 1):
            table[i, j] = min(
                table[i - 1, j] + 1,
                table[i, j - 1] + 1,
                table[i - 1, j - 1] + (0 if hyp[i - 1] == ref[j - 1] else 1)
            )

    return table[-1, -1]

def calculate_wer(hyp: list[str], ref: list[str]):
    word_error_count = wer_tabular(hyp, ref)
    return word_error_count / len(ref)

class CitationError(Exception):
    pass

def cite_passage(passage: str, patent: parser.ParsedPatent) -> tuple[parser.PatentTextLine, parser.PatentTextLine]:
    """
    Given a serialized passage that came from the document, convert the passage back into a list of parser.PatentTextLine objects, then return the two PatentTextLine objects marking the beginning and end of the passage, inclusive on both ends.

    The serialized specification should have been constructed using parser.serialized_specification_from_parsed_patent which returns a list of strings and then joined with "\n" to create a string. So splitting the selected passage on should recover the list of strings that can then be matched against the list of patent text lines in the parsed patent.
    """
    patent_text_lines = patent.text_lines

    # Split the passage into lines
    # passage_lines = passage.split('\n')
    # if len(passage_lines) > 1:
    #     def find_matching_line(line: str) -> parser.PatentTextLine | None:
    #         for patent_text_line in patent_text_lines.column_0:
    #             # This is kind of a hack and probably not very robust
    #             if patent_text_line.text.startswith(line):
    #                 return patent_text_line
    #         for patent_text_line in patent_text_lines.column_1:
    #             # This is kind of a hack and probably not very robust
    #             if patent_text_line.text.startswith(line):
    #                 return patent_text_line
    #         # Currently shouldn't get here, but just in case
    #         for patent_text_line in patent_text_lines.noncolumnar:
    #             # This is kind of a hack and probably not very robust
    #             if patent_text_line.text.startswith(line):
    #                 return patent_text_line
    #         return None
        
    #     start_line = find_matching_line(passage_lines[0])

    #     if start_line is None:
    #         raise CitationError(f"Could not find start line for passage: {passage}")
        
    #     end_line = find_matching_line(passage_lines[-1])

    #     if end_line is None:
    #         raise CitationError(f"Could not find end line for passage: {passage}")
        
    #     return (start_line, end_line)
    # else:

    best_passage = []
    best_wer = 1.0
    logger.info(f'Citing passage: {passage}')
    passage_words = passage.split()
    patent_words = parser.flatten_parsed_patent_to_words(patent)
    for i, patent_word in enumerate(patent_words):
        # We assume the first word in the passage returned by the language model will always be a well-formed word that exists in the patent. I.e., it won't be a word fragment, like the hyphenated part of a word. If it is, it probably matches a word in the parsed patent anyway because that's probably where the language model got that from.
        if patent_word.text == passage_words[0]:
            candidate_passage = patent_words[i:i+len(passage_words)]
            wer = calculate_wer([w.text for w in candidate_passage], passage_words)
            if wer < best_wer:
                best_wer = wer
                best_passage = candidate_passage
    if best_passage:
        return (best_passage[0].line, best_passage[-1].line)
    
    raise CitationError(f"Could not find start line for passage: {passage}")


import requests
import pathlib
import time
from logging import getLogger

from bs4 import BeautifulSoup

logger = getLogger(__name__)

# from patent_chart import parser

def google_request_with_retry(url, max_retries=3):
    for i in range(1, max_retries + 1):
        response = requests.get(url)
        if response.status_code == 200:
            return response
        else:
            logger.warning('Failed to request %s. Response code: %s. Error: %s', url, response.status_code, response.text)
            logger.warning('Retrying %s', url)
            time.sleep(i**2)
    return None

def parse_google_patent(patent_unique_id) -> tuple[str, str] | None:
    url = f"https://patents.google.com/patent/{patent_unique_id}/en"

    logger.info('Requesting %s', url)

    # Send an HTTP GET request to fetch the webpage
    response = google_request_with_retry(url)

    # Parse the HTML content of the page using BeautifulSoup
    soup = BeautifulSoup(response.text, 'html.parser')

    specification = soup.find('section', {'itemprop': 'description'})
    claims = soup.find('section', {'itemprop': 'claims'})

    if specification is None or claims is None:
        logger.warning('Failed to parse %s', url)
        return None

    return specification.text, claims.text

# if __name__ == '__main__':
#     package_dir = pathlib.Path(__file__).parents[1]
#     us_patents_dir = package_dir / 'us_patents_1980-2020'
#     for pdf_path in us_patents_dir.glob('*.pdf'):
#         unique_id = parser.parse_patent_unique_id_from_pdf_path(pdf_path)
#         if unique_id is not None:
#             google_unique_id = f'{unique_id.country_code}{unique_id.patent_number}{unique_id.kind_code or ""}'
#             result = parse_google_patent(google_unique_id)
#             if result is not None:
#                 specification, claims = result
#                 assert specification != ''
#                 assert claims != ''
#             else:
#                 print(f'Failed to parse {unique_id}')
#         else:
#             print(f'Failed to parse {pdf_path}')


import math
from typing import Callable

import numpy as np


def lcs_length(seq1: list[str], seq2: list[str]):
    """
    Longest common substring length
    """
    L = np.array([[0 for _ in range(len(seq2) + 1)] for _ in range(len(seq1) + 1)])
    z = 0
    for i in range(1, len(seq1) + 1):
        for j in range(1, len(seq2) + 1):
            if seq1[i - 1] == seq2[j - 1]:
                L[i, j] = L[i - 1, j - 1] + 1
                if L[i, j] > z:
                    z = L[i, j]
            else:
                L[i, j] = 0

    return z

def lcs(seq1: list[str], seq2: list[str]):
    """
    Longest common substring
    """
    L = np.array([[0 for _ in range(len(seq2) + 1)] for _ in range(len(seq1) + 1)])
    lcs_s: set[tuple[str, ...]] = set()
    z = 0
    for i in range(1, len(seq1) + 1):
        for j in range(1, len(seq2) + 1):
            if seq1[i - 1] == seq2[j - 1]:
                L[i, j] = L[i - 1, j - 1] + 1
                if L[i, j] > z:
                    z = L[i, j]
                    lcs_s.clear()
                    lcs_s.add(tuple(seq1[i - z:i]))
                elif L[i, j] == z:
                    lcs_s.add(tuple(seq1[i - z:i]))
            else:
                L[i, j] = 0


    return lcs_s

def default_hit_predicate(hyp: list[str], ref: list[str], hit_length: int = 10):
    return lcs_length(hyp, ref) > min(hit_length, len(ref))

def calculate_hit_rate(claim_element: tuple[float, str], selected_passages: list[str], reference: dict[tuple[float, str], list[str]], hit_predicate: Callable = default_hit_predicate):
    claim_element_text, reference_passages = reference.get(claim_element[0])

    selected_passages = [i.strip().split() for i in selected_passages]
    reference_passages = [i[0].strip().split() for i in reference_passages]

    hits = 0
    for reference_passage in reference_passages:
        for selected_passage in selected_passages:
            if hit_predicate(selected_passage, reference_passage):
                hits += 1
                break

    return hits / len(reference_passages)

def get_matching_passages(claim_element: tuple[float, str], selected_passages: list[str], reference: dict[tuple[float, str], list[str]], hit_predicate: Callable = default_hit_predicate):
    claim_element_text, reference_passages = reference.get(claim_element[0])
    selected_passages = [i.strip().split() for i in selected_passages]
    reference_passages = [i[0].strip().split() for i in reference_passages]

    matching_passages = []
    for reference_passage in reference_passages:
        for selected_passage in selected_passages:
            if hit_predicate(selected_passage, reference_passage):
                matching_passages.append((selected_passage, reference_passage))
                break
    
    return matching_passages


import pickle
from dataclasses import dataclass
import numpy as np

from sentence_transformers import SentenceTransformer

from patent_chart.parser import PatentTextLine, serialize_chunk

def get_model(model_name):
    if model_name == 'instructor-xl':
        return SentenceTransformer('hkunlp/instructor-xl')
    elif model_name == 'all-MiniLM-L6-v2':
        return SentenceTransformer('all-MiniLM-L6-v2')

def load_embeddings_from_pickle_dump(embeddings_path):
    with open(embeddings_path, "rb") as fIn:
        return pickle.load(fIn)
    
def save_spec_chunk_embeddings(spec_chunk_embeddings, embeddings_path):
    with open(embeddings_path, "wb") as fOut:
        pickle.dump(spec_chunk_embeddings, fOut, protocol=pickle.HIGHEST_PROTOCOL)

def cosine_similarity(a, b):
    return np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))

def find_most_similar_chunks(embeddings, summary_plus_claim_embedding, n=10):
    similarities = []
    for embedding in embeddings:
        similarities.append(cosine_similarity(summary_plus_claim_embedding, embedding))

    similarities = np.vstack(similarities).reshape(-1)
    return np.argsort(similarities)[-n:]

@dataclass
class EmbeddedSpec:
    paragraphs: list[list[PatentTextLine]]
    embeddings: list[np.ndarray]

def select_passages_for_claim_element(patent_summary_embedding: np.ndarray, patent_claim_element_embedding: np.ndarray, prior_art_spec: EmbeddedSpec, n_passages=10):
    claim_element_embedding = patent_claim_element_embedding.reshape(1, -1)
    most_similar_chunk_indices = find_most_similar_chunks(prior_art_spec.embeddings, claim_element_embedding + patent_summary_embedding, n=n_passages)
    selected_passages = []
    for i in most_similar_chunk_indices:
        selected_passages.append(serialize_chunk(prior_art_spec.paragraphs[i]))

    return selected_passages